import pandas as pd

from vivarium.framework.engine import Builder


class TreatmentIntervention:

    configuration_defaults = {
        "intervention": {
            "effect_size": 0.5,
        }
    }

    def __init__(self, name: str, affected_value: str):
        self.name = name
        self.affected_value = affected_value
        self.configuration_defaults = {
            name: TreatmentIntervention.configuration_defaults["intervention"]
        }

    # noinspection PyAttributeOutsideInit
    def setup(self, builder: Builder):
        effect_size = builder.configuration[self.name].effect_size
        builder.value.register_value_modifier(
            self.affected_value, modifier=self.intervention_effect
        )
        self.effect_size = builder.value.register_value_producer(
            f"{self.name}.effect_size",
            source=lambda index: pd.Series(effect_size, index=index),
        )

    def intervention_effect(self, index, value):
        return value * (1 - self.effect_size(index))



import numpy as np
import pandas as pd

from vivarium import InteractiveContext
from vivarium.framework.randomness import choice
from vivarium.framework.state_machine import Machine, State, Transition


def _population_fixture(column, initial_value):
    class PopFixture:
        @property
        def name(self):
            return f"test_pop_fixture_{column}_{initial_value}"

        def setup(self, builder):
            self.population_view = builder.population.get_view([column])
            builder.population.initializes_simulants(self.inner, creates_columns=[column])

        def inner(self, pop_data):
            self.population_view.update(pd.Series(initial_value, index=pop_data.index))

    return PopFixture()


def _even_population_fixture(column, values):
    class pop_fixture:
        @property
        def name(self):
            return "test_pop_fixture"

        def setup(self, builder):
            self.population_view = builder.population.get_view([column])
            builder.population.initializes_simulants(self.inner, creates_columns=[column])

        def inner(self, pop_data):
            self.population_view.update(choice("start", pop_data.index, values))

    return pop_fixture()


def test_transition():
    done_state = State("done")
    start_state = State("start")
    start_state.add_transition(done_state)
    machine = Machine("state", states=[start_state, done_state])

    simulation = InteractiveContext(
        components=[machine, _population_fixture("state", "start")]
    )
    event_time = simulation._clock.time + simulation._clock.step_size
    machine.transition(simulation.get_population().index, event_time)
    assert np.all(simulation.get_population().state == "done")


def test_choice(base_config):
    base_config.update({"population": {"population_size": 10000}})
    a_state = State("a")
    b_state = State("b")
    start_state = State("start")
    start_state.add_transition(
        a_state, probability_func=lambda agents: np.full(len(agents), 0.5)
    )
    start_state.add_transition(
        b_state, probability_func=lambda agents: np.full(len(agents), 0.5)
    )
    machine = Machine("state", states=[start_state, a_state, b_state])

    simulation = InteractiveContext(
        components=[machine, _population_fixture("state", "start")], configuration=base_config
    )
    event_time = simulation._clock.time + simulation._clock.step_size
    machine.transition(simulation.get_population().index, event_time)
    a_count = (simulation.get_population().state == "a").sum()
    assert round(a_count / len(simulation.get_population()), 1) == 0.5


def test_null_transition(base_config):
    base_config.update({"population": {"population_size": 10000}})
    a_state = State("a")
    start_state = State("start")
    start_state.add_transition(
        a_state, probability_func=lambda agents: np.full(len(agents), 0.5)
    )
    start_state.allow_self_transitions()

    machine = Machine("state", states=[start_state, a_state])

    simulation = InteractiveContext(
        components=[machine, _population_fixture("state", "start")], configuration=base_config
    )
    event_time = simulation._clock.time + simulation._clock.step_size
    machine.transition(simulation.get_population().index, event_time)
    a_count = (simulation.get_population().state == "a").sum()
    assert round(a_count / len(simulation.get_population()), 1) == 0.5


def test_no_null_transition(base_config):
    base_config.update({"population": {"population_size": 10000}})
    a_state = State("a")
    b_state = State("b")
    start_state = State("start")
    a_transition = Transition(
        start_state, a_state, probability_func=lambda index: pd.Series(0.5, index=index)
    )
    b_transition = Transition(
        start_state, b_state, probability_func=lambda index: pd.Series(0.5, index=index)
    )
    start_state.transition_set.allow_null_transition = False
    start_state.transition_set.extend((a_transition, b_transition))
    machine = Machine("state")
    machine.states.extend([start_state, a_state, b_state])

    simulation = InteractiveContext(
        components=[machine, _population_fixture("state", "start")], configuration=base_config
    )
    event_time = simulation._clock.time + simulation._clock.step_size
    machine.transition(simulation.get_population().index, event_time)
    a_count = (simulation.get_population().state == "a").sum()
    assert round(a_count / len(simulation.get_population()), 1) == 0.5


def test_side_effects():
    class DoneState(State):
        @property
        def name(self):
            return "test_done_state"

        def setup(self, builder):
            super().setup(builder)
            self.population_view = builder.population.get_view(["count"])

        def _transition_side_effect(self, index, event_time):
            pop = self.population_view.get(index)
            self.population_view.update(pop["count"] + 1)

    done_state = DoneState("done")
    start_state = State("start")
    start_state.add_transition(done_state)
    done_state.add_transition(start_state)

    machine = Machine("state", states=[start_state, done_state])

    simulation = InteractiveContext(
        components=[
            machine,
            _population_fixture("state", "start"),
            _population_fixture("count", 0),
        ]
    )
    event_time = simulation._clock.time + simulation._clock.step_size
    machine.transition(simulation.get_population().index, event_time)
    assert np.all(simulation.get_population()["count"] == 1)
    machine.transition(simulation.get_population().index, event_time)
    assert np.all(simulation.get_population()["count"] == 1)
    machine.transition(simulation.get_population().index, event_time)
    assert np.all(simulation.get_population()["count"] == 2)



"""
===================
Resource Management
===================

This module provides a tool to manage dependencies on resources within a
:mod:`vivarium` simulation. These resources take the form of things that can
be created and utilized by components, for example columns in the
:mod:`state table <vivarium.framework.population>`
or :mod:`named value pipelines <vivarium.framework.values>`.

Because these resources need to be created before they can be used, they are
sensitive to ordering. The intent behind this tool is to provide an interface
that allows other managers to register resources with the resource manager
and in turn ask for ordered sequences of these resources according to their
dependencies or raise exceptions if this is not possible.

"""
from types import MethodType
from typing import Any, Callable, Iterable, List

import networkx as nx
from loguru import logger

from vivarium.exceptions import VivariumError


class ResourceError(VivariumError):
    """Error raised when a dependency requirement is violated."""

    pass


RESOURCE_TYPES = {
    "value",
    "value_source",
    "missing_value_source",
    "value_modifier",
    "column",
    "stream",
}
NULL_RESOURCE_TYPE = "null"


class ResourceGroup:
    """Resource groups are the nodes in the resource dependency graph.

    A resource group represents the pool of resources produced by a single
    callable and all the dependencies necessary to produce that resource.
    When thinking of the dependency graph, this represents a vertex and
    all in-edges.  This is a local-information representation that can be
    used to construct the entire dependency graph once all resources are
    specified.

    """

    def __init__(
        self,
        resource_type: str,
        resource_names: List[str],
        producer: Callable,
        dependencies: List[str],
    ):
        self._resource_type = resource_type
        self._resource_names = resource_names
        self._producer = producer
        self._dependencies = dependencies

    @property
    def type(self) -> str:
        """The type of resource produced by this resource group's producer.

        Must be one of `RESOURCE_TYPES`.

        """
        return self._resource_type

    @property
    def names(self) -> List[str]:
        """The long names (including type) of all resources in this group."""
        return [f"{self._resource_type}.{name}" for name in self._resource_names]

    @property
    def producer(self) -> Any:
        """The method or object that produces this group of resources."""
        return self._producer

    @property
    def dependencies(self) -> List[str]:
        """The long names (including type) of dependencies for this group."""
        return self._dependencies

    def __iter__(self) -> Iterable[str]:
        return iter(self.names)

    def __repr__(self) -> str:
        resources = ", ".join(self)
        return f"ResourceProducer({resources})"

    def __str__(self) -> str:
        resources = ", ".join(self)
        return f"({resources})"


class ResourceManager:
    """Manages all the resources needed for population initialization."""

    def __init__(self):
        # This will be a dict with string keys representing the the resource
        # and the resource group they belong to. This is a one to many mapping
        # as some resource groups contain many resources.
        self._resource_group_map = {}
        # null producers are those that don't produce any resources externally
        # but still consume other resources (i.e., have dependencies) - these
        # are only pop initializers as of 9/26/2019. Tracker is here to assign
        # them unique ids.
        self._null_producer_count = 0
        # Attribute used for lazy (but cached) graph initialization.
        self._graph = None
        # Attribute used for lazy (but cached) graph topo sort.
        self._sorted_nodes = None

    @property
    def name(self) -> str:
        """The name of this manager."""
        return "resource_manager"

    @property
    def graph(self) -> nx.DiGraph:
        """The networkx graph representation of the resource pool."""
        if self._graph is None:
            self._graph = self._to_graph()
        return self._graph

    @property
    def sorted_nodes(self):
        """Returns a topological sort of the resource graph.

        Notes
        -----
        Topological sorts are not stable. Be wary of depending on order
        where you shouldn't.

        """
        if self._sorted_nodes is None:
            try:
                self._sorted_nodes = list(nx.algorithms.topological_sort(self.graph))
            except nx.NetworkXUnfeasible:
                raise ResourceError(
                    f"The resource pool contains at least one cycle: "
                    f"{nx.find_cycle(self.graph)}."
                )
        return self._sorted_nodes

    def add_resources(
        self,
        resource_type: str,
        resource_names: List[str],
        producer: Any,
        dependencies: List[str],
    ):
        """Adds managed resources to the resource pool.

        Parameters
        ----------
        resource_type
            The type of the resources being added. Must be one of
            `RESOURCE_TYPES`.
        resource_names
            A list of names of the resources being added.
        producer
            A method or object that will produce the resources.
        dependencies
            A list of resource names formatted as
            ``resource_type.resource_name`` that the producer requires.

        Raises
        ------
        ResourceError
            If either the resource type is invalid, a component has multiple
            resource producers for the ``column`` resource type, or
            there are multiple producers of the same resource.

        """
        if resource_type not in RESOURCE_TYPES:
            raise ResourceError(
                f"Unknown resource type {resource_type}. "
                f"Permitted types are {RESOURCE_TYPES}."
            )

        resource_group = self._get_resource_group(
            resource_type, resource_names, producer, dependencies
        )

        for resource in resource_group:
            if resource in self._resource_group_map:
                other_producer = self._resource_group_map[resource].producer
                raise ResourceError(
                    f"Both {producer} and {other_producer} are registered as "
                    f"producers for {resource}."
                )
            self._resource_group_map[resource] = resource_group

    def _get_resource_group(
        self,
        resource_type: str,
        resource_names: List[str],
        producer: MethodType,
        dependencies: List[str],
    ) -> ResourceGroup:
        """Packages resource information into a resource group.

        See Also
        --------
        :class:`ResourceGroup`

        """
        if not resource_names:
            # We have a "producer" that doesn't produce anything, but
            # does have dependencies. This is necessary for components that
            # want to track private state information.
            resource_type = NULL_RESOURCE_TYPE
            resource_names = [str(self._null_producer_count)]
            self._null_producer_count += 1

        return ResourceGroup(resource_type, resource_names, producer, dependencies)

    def _to_graph(self) -> nx.DiGraph:
        """Constructs the full resource graph from information in the groups.

        Components specify local dependency information during setup time.
        When the resources are required at population creation time,
        the graph is generated as all resources must be registered at that
        point.

        Notes
        -----
        We are taking advantage of lazy initialization to sneak this in
        between post setup time when the :class:`values manager
        <vivarium.framework.values.ValuesManager>` finalizes pipeline
        dependencies and population creation time.

        """
        resource_graph = nx.DiGraph()
        # networkx ignores duplicates
        resource_graph.add_nodes_from(self._resource_group_map.values())

        for resource_group in resource_graph.nodes:
            for dependency in resource_group.dependencies:
                if dependency not in self._resource_group_map:
                    # Warn here because this sometimes happens naturally
                    # if observer components are missing from a simulation.
                    logger.warning(
                        f"Resource {dependency} is not provided by any component but is needed to "
                        f"compute {resource_group}."
                    )
                    continue
                dependency_group = self._resource_group_map[dependency]
                resource_graph.add_edge(dependency_group, resource_group)

        return resource_graph

    def __iter__(self) -> Iterable[MethodType]:
        """Returns a dependency-sorted iterable of population initializers.

        We exclude all non-initializer dependencies. They were necessary in
        graph construction, but we only need the column producers at population
        creation time.

        """
        return iter(
            [
                r.producer
                for r in self.sorted_nodes
                if r.type in {"column", NULL_RESOURCE_TYPE}
            ]
        )

    def __repr__(self):
        out = {}
        for resource_group in set(self._resource_group_map.values()):
            produced = ", ".join(resource_group)
            out[produced] = ", ".join(resource_group.dependencies)
        return "\n".join([f"{produced} : {depends}" for produced, depends in out.items()])


class ResourceInterface:
    """The resource management system.

    A resource in :mod:`vivarium` is something like a state table column
    or a randomness stream. These resources are used to initialize or alter
    the state of the simulation. Many of these resources might depend on each
    other and therefore need to be created or updated in a particular order.
    These dependency chains can be quite long and complex.

    Placing the ordering responsibility on end users makes simulations very
    fragile and difficult to understand. Instead, the resource management
    system allows users to only specify local dependencies. The system then
    uses the local dependency information to construct a full dependency
    graph, validate that there are no cyclic dependencies, and return
    resources and their producers in an order that makes sense.

    """

    def __init__(self, manager: ResourceManager):
        self._manager = manager

    def add_resources(
        self,
        resource_type: str,
        resource_names: List[str],
        producer: Any,
        dependencies: List[str],
    ):
        """Adds managed resources to the resource pool.

        Parameters
        ----------
        resource_type
            The type of the resources being added. Must be one of
            `RESOURCE_TYPES`.
        resource_names
            A list of names of the resources being added.
        producer
            A method or object that will produce the resources.
        dependencies
            A list of resource names formatted as
            ``resource_type.resource_name`` that the producer requires.

        Raises
        ------
        ResourceError
            If either the resource type is invalid, a component has multiple
            resource producers for the ``column`` resource type, or
            there are multiple producers of the same resource.

        """
        self._manager.add_resources(resource_type, resource_names, producer, dependencies)

    def __iter__(self):
        """Returns a dependency-sorted iterable of population initializers.

        We exclude all non-initializer dependencies. They were necessary in
        graph construction, but we only need the column producers at population
        creation time.

        """
        return iter(self._manager)



from typing import Callable, List

import pandas as pd


class ResultsManager:
    @property
    def name(self):
        return "results_manager"

    def add_mapping_strategy(self, *args, **kwargs):
        ...

    def add_default_grouping_columns(self, *args, **kwargs):
        ...

    def register_results_producer(self, *args, **kwargs):
        ...


class ResultsInterface:
    """Builder interface for the results management system.

    The results management system allows users to delegate results production
    to the simulation framework. This process attempts to roughly mimic the
    groupby-apply logic commonly done when manipulating :mod:`pandas`
    DataFrames. The representation of state in the simulation is complex,
    however, as it includes information both in the population state table
    as well as dynamically generated information available from the
    :class:`value pipelines <vivarium.framework.values.Pipeline>`.
    Additionally, good encapsulation of simulation logic typically has
    results production separated from the modeling code into specialized
    `Observer` components. This often highlights the need for transformations
    of the simulation state into representations that aren't needed for
    modeling, but are required for the stratification of produced results.

    To support these complexities, we first allow the user to supplement the
    simulation state with transformations of the underlying state table and
    pipeline values using :meth:`add_mapping_strategy`. For example, a public
    health simulation may want to group its summary results by 5-year age
    bins. A user could add a mapping strategy to transform the continuous
    "age" property of each of the simulants into a discrete set of age
    categories. This expanded version of the state can then be used to
    stratify the population into groups of interest for producing aggregate
    statistics.

    In general, most of this grouping is for all results being produced.  E.g.
    if we're going to stratify results in a public health simulation by age
    group and/or sex, we likely want to do it for all results and not just a
    few. These global stratifications can be registered with a call to
    :meth:`add_default_grouping_columns` where the columns in question are part
    of the expanded state table generated by the previously discussed mapping
    strategies.

    The suggested pattern for adding mapping strategies and default
    grouping columns is to centralize them in the `setup` method of a general
    `Metrics` or `Results` component. From this component you can expose
    configuration for the global stratification of results. This component
    should be fairly easy to reuse across entire categories of models.

    Individual `Observer` components (or model components if modeling and
    results production are mixed in the same component) can then declare
    strategies to produce result measures using
    :meth:`register_results_producer`. This method is intended to closely
    mirror the `apply` step of a :mod:`pandas` groupby-apply, taking a generic
    callable that operates on dataframes with the understanding that it
    will be called on data subsets generated by groupings defined by
    :meth:`add_default_grouping_columns`. For flexibility, a user can
    elect to add additional stratification or exclude some of the default
    stratification levels when registering a results producer.

    Results producers registered by users should only produce results about
    the current time step. The results system automatically accumulates the
    results across the whole lifetime of the simulation. The combination of the
    'measure' used when registering a results producer and the levels of
    the stratification groupings form a unique key for a particular result
    and calls to produce results every time step will accumulate values
    for a particular result if it is generated multiple times.  For example,
    we may want to track all the heart attacks that occur for 50-55 year old
    men in a several year simulation of cardiovascular health. Each time step,
    the results system will call the producer the user has registered to count
    heart attacks by age group and that producer will count all the heart
    attacks in that particular time step. The results system will then add
    that time step's worth of heart attacks for 50-55 year old men to its
    running total.

    """

    def __init__(self, manager: ResultsManager) -> None:
        self._manager = manager

    def add_mapping_strategy(
        self, new_column: str, mapper: Callable[[pd.Index], pd.Series]
    ) -> None:
        """Adds a specification to map simulation state into a new column.

        This allows a user to specify an arbitrary function to generate
        new columns in the expanded state table for use in results production.

        For instance, suppose you have a column `time_of_death` that records
        the time a simulant dies as a time stamp.  In order to stratify
        results, you might map that column to a `year_of_death` column to
        group deaths by the year in which they occurred.

        Parameters
        ----------
        new_column
            The name of the new column in the expanded state table.
        mapper
            A callable that takes an index representing the entire
            population and returning a series of values indexed by
            the population index.

        """
        self._manager.add_mapping_strategy(new_column, mapper)

    def add_default_grouping_columns(self, grouping_columns: List[str]) -> None:
        """Add a list of expanded state table columns to group by for all measures.

        Generally, we want to be able to globally define the stratification for all
        results produced by a simulation. This method allows a user to provide
        a set of expanded state table columns that will be used to group the population
        for all results production strategies. These defaults can be
        overridden by individual results production strategies for special cases.

        Parameters
        ----------
            grouping_columns
                A list of names of columns in the expanded state table to use
                when grouping the population for results production.

        """
        self._manager.add_default_grouping_columns(grouping_columns)

    def add_results_production_strategy(
        self,
        measure: str,
        pop_filter: str = "",
        aggregator: Callable[[pd.DataFrame], float] = len,
        additional_grouping_columns: List[str] = (),
        excluded_grouping_columns: List[str] = (),
        when: str = "collect_metrics",
    ) -> None:
        """Provide the framework with a strategy for producing a results 'measure'.

        Results production strategies operate like functions used in the
        `apply` step of a :mod:`pandas` groupby-apply.

        Parameters
        ----------
        measure
            The name of the measure to be produced by the strategy.
        pop_filter
            A filter to apply to the population before grouping and
            aggregation. Filters should be formatted so that they are
            consistent with arguments to :func:`pandas.DataFrame.query`.
        aggregator
            A callable operating on a :obj:`pandas.DataFrame` and producing
            a float value. Defaults to the length of the data, producing
            a group count.
        additional_grouping_columns
            Columns in the expanded state table to group the population
            by before applying any aggregation. These columns are added
            to any default grouping columns registered with
            :func:`ResultsInterface.add_default_grouping_columns`.
        excluded_grouping_columns
            Columns in the default grouping columns registered with
            :func:`ResultsInterface.add_default_grouping_columns` that should
            not be used in the production of this measure.
        when
            The name of the time step event when this measure should be
            produced.

        """
        self._manager.register_results_producer(
            measure,
            pop_filter,
            aggregator,
            list(additional_grouping_columns),
            list(excluded_grouping_columns),
            when,
        )



"""
==========================
Vivarium Interactive Tools
==========================

This module provides an interface for interactive simulation usage. The main
part is the :class:`InteractiveContext`, a sub-class of the main simulation
object in ``vivarium`` that has been extended to include convenience
methods for running and exploring the simulation in an interactive setting.

See the associated tutorials for :ref:`running <interactive_tutorial>` and
:ref:`exploring <exploration_tutorial>` for more information.

"""
from math import ceil
from typing import Any, Callable, Dict, List

import pandas as pd

from vivarium.framework.engine import SimulationContext
from vivarium.framework.time import Time, Timedelta
from vivarium.framework.values import Pipeline

from .utilities import log_progress, run_from_ipython


class InteractiveContext(SimulationContext):
    """A simulation context with helper methods for running simulations
    interactively.

    This class should not be instantiated directly. It should be created with a
    call to one of the helper methods provided in this module.
    """

    def __init__(self, *args, setup=True, **kwargs):
        super().__init__(*args, **kwargs)

        if setup:
            self.setup()

    @property
    def current_time(self) -> Time:
        """Returns the current simulation time."""
        return self._clock.time

    def setup(self):
        super().setup()
        self.initialize_simulants()

    def step(self, step_size: Timedelta = None):
        """Advance the simulation one step.

        Parameters
        ----------
        step_size
            An optional size of step to take. Must be the same type as the
            simulation clock's step size (usually a pandas.Timedelta).
        """
        old_step_size = self._clock.step_size
        if step_size is not None:
            if not isinstance(step_size, type(self._clock.step_size)):
                raise ValueError(
                    f"Provided time must be an instance of {type(self._clock.step_size)}"
                )
            self._clock._step_size = step_size
        super().step()
        self._clock._step_size = old_step_size

    def run(self, with_logging: bool = True) -> int:
        """Run the simulation for the duration specified in the configuration.

        Parameters
        ----------
        with_logging
            Whether or not to log the simulation steps. Only works in an ipython
            environment.

        Returns
        -------
        int
            The number of steps the simulation took.

        """
        return self.run_until(self._clock.stop_time, with_logging=with_logging)

    def run_for(self, duration: Timedelta, with_logging: bool = True) -> int:
        """Run the simulation for the given time duration.

        Parameters
        ----------
        duration
            The length of time to run the simulation for. Should be the same
            type as the simulation clock's step size (usually a pandas
            Timedelta).
        with_logging
            Whether or not to log the simulation steps. Only works in an ipython
            environment.

        Returns
        -------
        int
            The number of steps the simulation took.

        """
        return self.run_until(self._clock.time + duration, with_logging=with_logging)

    def run_until(self, end_time: Time, with_logging: bool = True) -> int:
        """Run the simulation until the provided end time.

        Parameters
        ----------
        end_time
            The time to run the simulation until. The simulation will run until
            its clock is greater than or equal to the provided end time.
        with_logging
            Whether or not to log the simulation steps. Only works in an ipython
            environment.

        Returns
        -------
        int
            The number of steps the simulation took.

        """
        if not isinstance(end_time, type(self._clock.time)):
            raise ValueError(f"Provided time must be an instance of {type(self._clock.time)}")

        iterations = int(ceil((end_time - self._clock.time) / self._clock.step_size))
        self.take_steps(number_of_steps=iterations, with_logging=with_logging)
        assert self._clock.time - self._clock.step_size < end_time <= self._clock.time
        return iterations

    def take_steps(
        self, number_of_steps: int = 1, step_size: Timedelta = None, with_logging: bool = True
    ):
        """Run the simulation for the given number of steps.

        Parameters
        ----------
        number_of_steps
            The number of steps to take.
        step_size
            An optional size of step to take. Must be the same type as the
            simulation clock's step size (usually a pandas.Timedelta).
        with_logging
            Whether or not to log the simulation steps. Only works in an ipython
            environment.

        """
        if not isinstance(number_of_steps, int):
            raise ValueError("Number of steps must be an integer.")

        if run_from_ipython() and with_logging:
            for _ in log_progress(range(number_of_steps), name="Step"):
                self.step(step_size)
        else:
            for _ in range(number_of_steps):
                self.step(step_size)

    def get_population(self, untracked: bool = False) -> pd.DataFrame:
        """Get a copy of the population state table.

        Parameters
        ----------
        untracked
            Whether or not to return simulants who are no longer being tracked
            by the simulation.

        """
        return self._population.get_population(untracked)

    def list_values(self) -> List[str]:
        """List the names of all pipelines in the simulation."""
        return list(self._values.keys())

    def get_value(self, value_pipeline_name: str) -> Pipeline:
        """Get the value pipeline associated with the given name."""
        return self._values.get_value(value_pipeline_name)

    def list_events(self) -> List[str]:
        """List all event types registered with the simulation."""
        return self._events.list_events()

    def get_listeners(self, event_type: str) -> List[Callable]:
        """Get all listeners of a particular type of event.

        Available event types can be found by calling
        :func:`InteractiveContext.list_events`.

        Parameters
        ----------
        event_type
            The type of event to grab the listeners for.

        """
        if event_type not in self._events:
            raise ValueError(f"No event {event_type} in system.")
        return self._events.get_listeners(event_type)

    def get_emitter(self, event_type: str) -> Callable:
        """Get the callable that emits the given type of events.

        Available event types can be found by calling
        :func:`InteractiveContext.list_events`.

        Parameters
        ----------
        event_type
            The type of event to grab the listeners for.

        """
        if event_type not in self._events:
            raise ValueError(f"No event {event_type} in system.")
        return self._events.get_emitter(event_type)

    def list_components(self) -> Dict[str, Any]:
        """Get a mapping of component names to components currently in the simulation.

        Returns
        -------
        Dict[str, Any]
            A dictionary mapping component names to components.

        """
        return self._component_manager.list_components()

    def get_component(self, name: str) -> Any:
        """Get the component in the simulation that has ``name``, if present.
        Names are guaranteed to be unique.

        Parameters
        ----------
        name
            A component name.
        Returns
        -------
            A component that has the name ``name`` else None.

        """
        return self._component_manager.get_component(name)

    def __repr__(self):
        return "InteractiveContext()"



from .disease import DiseaseModel, DiseaseState, DiseaseTransition, SISDiseaseModel
from .intervention import TreatmentIntervention
from .mortality import Mortality
from .observer import Observer
from .population import BasePopulation
from .risk import Risk, RiskEffect


def get_model_specification_path():
    from pathlib import Path

    p = Path(__file__).parent / "disease_model.yaml"
    return str(p)


def get_disease_model_simulation():
    from vivarium import InteractiveContext

    p = get_model_specification_path()
    return InteractiveContext(p)



"""
=====================
Life Cycle Management
=====================

The life cycle is a representation of the flow of execution states in a
:mod:`vivarium` simulation. The tools in this model allow a simulation to
formally represent its execution state and use the formal representation to
enforce run-time contracts.

There are two flavors of contracts that this system enforces:

 - **Constraints**: These are contracts around when certain methods,
   particularly those available off the :ref:`Builder <builder_concept>`,
   can be used. For example, :term:`simulants <Simulant>` should only be
   added to the simulation during initial population creation and during
   the main simulation loop, otherwise services necessary for initializing
   that population's attributes may not exist. By applying a constraint,
   we can provide very clear errors about what went wrong, rather than
   a deep and unintelligible stack trace.
 - **Ordering Contracts**: The
   :class:`~vivarium.framework.engine.SimulationContext` will construct
   the formal representation of the life cycle during its initialization.
   Once generated, the context declares as it transitions between
   different lifecycle states and the tools here ensure that only valid
   transitions occur.  These kinds of contracts are particularly useful
   during interactive usage, as they prevent users from, for example,
   running a simulation whose population has not been created.

The tools here also allow for introspection of the simulation life cycle.

"""
import functools
import textwrap
from typing import Callable, List, Optional, Tuple

from vivarium.exceptions import VivariumError


class LifeCycleError(VivariumError):
    """Generic error class for the life cycle management system."""

    pass


class InvalidTransitionError(LifeCycleError):
    """Error raised when life cycle ordering contracts are violated."""

    pass


class ConstraintError(LifeCycleError):
    """Error raised when life cycle constraint contracts are violated."""

    pass


class LifeCycleState:
    """A representation of a simulation run state."""

    def __init__(self, name: str):
        self._name = name
        self._next = None
        self._loop_next = None
        self._entrance_count = 0
        self._handlers = []

    @property
    def name(self) -> str:
        """The name of the lifecycle state."""
        return self._name

    @property
    def entrance_count(self) -> int:
        """The number of times this state has been entered."""
        return self._entrance_count

    def add_next(self, next_state: "LifeCycleState", loop: bool = False):
        """Link this state to the next state in the simulation life cycle.

        States are linked together and used to ensure that the simulation
        life cycle proceeds in the proper order.  A life cycle state can be
        bound to two ``next`` states to allow for loops in the life cycle and
        both are considered valid when checking for valid state transitions.
        The first represents the linear progression through the simulation,
        while the second represents a loop in the life cycle.

        Parameters
        ----------
        next_state
            The next state in the simulation life cycle.
        loop
            Whether the provided state is the linear next state or a loop
            back to a previous state in the life cycle.

        """
        if loop:
            self._loop_next = next_state
        else:
            self._next = next_state

    def valid_next_state(self, state: Optional["LifeCycleState"]) -> bool:
        """Check if the provided state is valid for a life cycle transition.

        Parameters
        ----------
        state
            The state to check.

        Returns
        -------
        bool
            Whether the state is valid for a transition.

        """
        return (state is None and state is self._next) or (
            state is not None and (state is self._next or state is self._loop_next)
        )

    def enter(self):
        """Marks an entrance into this state."""
        self._entrance_count += 1

    def add_handlers(self, handlers: List[Callable]):
        """Registers a set of functions that will be executed during the state.

        The primary use case here is for introspection and reporting.
        For setting constraints, see :meth:`LifeCycleInterface.add_constraint`.

        Parameters
        ----------
        handlers
            The set of functions that will be executed during this state.

        """
        for h in handlers:
            name = h.__name__
            if hasattr(h, "__self__"):
                obj = h.__self__
                self._handlers.append(f"{obj.__class__.__name__}({obj.name}).{name}")
            else:
                self._handlers.append(f"Unbound function {name}")

    def __repr__(self) -> str:
        return f"LifeCycleState(name={self.name})"

    def __str__(self) -> str:
        return "\n\t".join([self.name] + self._handlers)


class LifeCyclePhase:
    """A representation of a distinct lifecycle phase in the simulation.

    A lifecycle phase is composed of one or more unique lifecycle states.
    There is exactly one state within the phase which serves as a valid
    exit point from the phase.  The states may operate in a loop.

    """

    def __init__(self, name: str, states: List[str], loop: bool):
        self._name = name
        self._states = [LifeCycleState(states[0])]
        self._loop = loop
        for s in states[1:]:
            self._states.append(LifeCycleState(s))
            self._states[-2].add_next(self._states[-1])
        if self._loop:
            self._states[-1].add_next(self._states[0], loop=True)

    @property
    def name(self) -> str:
        """The name of this life cycle phase."""
        return self._name

    @property
    def states(self) -> Tuple[LifeCycleState]:
        """The states in this life cycle phase in order of execution."""
        return tuple(self._states)

    def add_next(self, phase: "LifeCyclePhase"):
        """Link the provided phase as the next phase in the life cycle."""
        self._states[-1].add_next(phase._states[0])

    def get_state(self, state_name: str) -> LifeCycleState:
        """Retrieve a life cycle state by name from the phase."""
        return [s for s in self._states if s.name == state_name].pop()

    def __contains__(self, state_name: str) -> bool:
        return bool([s for s in self._states if s.name == state_name])

    def __repr__(self) -> str:
        return f"LifeCyclePhase(name={self.name}, states={[s.name for s in self.states]})"

    def __str__(self) -> str:
        out = self.name
        if self._loop:
            out += "*"
        out += "\n" + textwrap.indent("\n".join([str(state) for state in self.states]), "\t")
        return out


class LifeCycle:
    """A concrete representation of the flow of simulation execution states."""

    def __init__(self):
        self._state_names = set()
        self._phase_names = set()
        self._phases = []
        self.add_phase("initialization", ["initialization"], loop=False)

    def add_phase(self, phase_name: str, states: List[str], loop):
        """Add a new phase to the lifecycle.

        Phases must be added in order.

        Parameters
        ----------
        phase_name
            The name of the phase to add.  Phase names must be unique.
        states
            The list of names (in order) of the states that make up the
            life cycle phase.  State names must be unique across the entire
            life cycle.
        loop
            Whether the life cycle phase states loop.

        Raises
        ------
        LifeCycleError
            If the phase or state names are non-unique.

        """
        self._validate(phase_name, states)

        new_phase = LifeCyclePhase(phase_name, states, loop)
        if self._phases:
            self._phases[-1].add_next(new_phase)

        self._state_names.update(states)
        self._phase_names.add(phase_name)
        self._phases.append(new_phase)

    def get_state(self, state_name: str) -> LifeCycleState:
        """Retrieve a life cycle state from the life cycle.

        Parameters
        ----------
        state_name
            The name of the state to retrieve

        Returns
        -------
        LifeCycleState
            The requested state.

        Raises
        ------
        LifeCycleError
            If the requested state does not exist.

        """
        if state_name not in self:
            raise LifeCycleError(f"Attempting to look up non-existent state {state_name}.")
        phase = [p for p in self._phases if state_name in p].pop()
        return phase.get_state(state_name)

    def get_state_names(self, phase_name: str) -> List[str]:
        """Retrieve the names of all states in the provided phase.

        Parameters
        ----------
        phase_name
            The name of the phase to retrieve the state names from.

        Return
        ------
        List[str]
            The state names in the provided phase.

        Raises
        ------
        LifeCycleError
            If the phase does not exist in the life cycle.

        """
        if phase_name not in self._phase_names:
            raise LifeCycleError(
                f"Attempting to look up states from non-existent phase {phase_name}."
            )
        phase = [p for p in self._phases if p.name == phase_name].pop()
        return [s.name for s in phase.states]

    def _validate(self, phase_name: str, states: List[str]):
        """Validates that a phase and set of states are unique."""
        if phase_name in self._phase_names:
            raise LifeCycleError(
                f"Lifecycle phase names must be unique. You're attempting "
                f"to add {phase_name} but it already exists."
            )
        if len(states) != len(set(states)):
            raise LifeCycleError(
                f"Attempting to create a life cycle phase with duplicate state names. "
                f"States: {states}"
            )
        duplicates = self._state_names.intersection(states)
        if duplicates:
            raise LifeCycleError(
                f"Lifecycle state names must be unique.  You're attempting "
                f"to add {duplicates} but they already exist."
            )

    def __contains__(self, state_name: str) -> bool:
        return state_name in self._state_names

    def __repr__(self) -> str:
        return f"LifeCycle(phases={self._phase_names})"

    def __str__(self) -> str:
        return "\n".join([str(phase) for phase in self._phases])


class ConstraintMaker:
    """Factory for making state-based constraints on component methods."""

    def __init__(self, lifecycle_manager):
        self.lifecycle_manager = lifecycle_manager
        self.constraints = set()

    def check_valid_state(self, method: Callable, permitted_states: List[str]):
        """Ensures a component method is being called during an allowed state.

        Parameters
        ----------
        method
            The method the constraint is applied to.
        permitted_states
            The states in which the method is permitted to be called.

        Raises
        ------
        ConstraintError
            If the method is being called outside the permitted states.

        """
        current_state = self.lifecycle_manager.current_state
        if current_state not in permitted_states:
            raise ConstraintError(
                f"Trying to call {method} during {current_state},"
                f" but it may only be called during {permitted_states}."
            )

    def constrain_normal_method(
        self, method: Callable, permitted_states: List[str]
    ) -> Callable:
        """Only permit a method to be called during the provided states.

        Constraints are applied by dynamically wrapping and binding a method
        to an existing component at run time.

        Parameters
        ----------
        method
            The method to constrain.
        permitted_states
            The life cycle states in which the method can be called.

        Returns
        -------
        Callable
            The constrained method.

        """

        @functools.wraps(method)
        def _wrapped(*args, **kwargs):
            self.check_valid_state(method, permitted_states)
            # Call the __func__ because we're rebinding _wrapped to the method
            # name on the object.  If we called method directly, we'd get
            # two copies of self.
            return method.__func__(*args, **kwargs)

        # Invoke the descriptor protocol to bind the wrapped method to the
        # component instance.
        rebound_method = _wrapped.__get__(method.__self__, method.__self__.__class__)
        # Then update the instance dictionary to reflect that the wrapped
        # method is bound to the original name.
        setattr(method.__self__, method.__name__, rebound_method)
        return rebound_method

    @staticmethod
    def to_guid(method: Callable) -> str:
        """Convert a method on to a global id.

        Because we dynamically rebind methods, the old ones will get garbage
        collected, making :func:`id` unreliable for checking if a method
        has been constrained before.

        """
        return f"{method.__self__.name}.{method.__name__}"

    def __call__(self, method: Callable, permitted_states: List[str]) -> Callable:
        """Only permit a method to be called during the provided states.

        Constraints are applied by dynamically wrapping and binding a method
        to an existing component at run time.

        Parameters
        ----------
        method
            The method to constrain.
        permitted_states
            The life cycle states in which the method can be called.

        Returns
        -------
            The constrained method.

        Raises
        ------
        TypeError
            If an unbound function is supplied for constraint.
        ValueError
            If the provided method is a python "special" method (i.e. a
            method surrounded by double underscores).

        """
        if not hasattr(method, "__self__"):
            raise TypeError(
                "Can only apply constraints to bound object methods. "
                f"You supplied the function {method}."
            )
        name = method.__name__
        if name.startswith("__") and name.endswith("__"):
            raise ValueError(
                "Can only apply constraints to normal object methods. "
                f" You supplied {method}."
            )

        if self.to_guid(method) in self.constraints:
            raise ConstraintError(f"Method {method} has already been constrained.")

        self.constraints.add(self.to_guid(method))
        return self.constrain_normal_method(method, permitted_states)


class LifeCycleManager:
    """Manages ordering- and constraint-based contracts in the simulation."""

    def __init__(self):
        self.lifecycle = LifeCycle()
        self._current_state = self.lifecycle.get_state("initialization")
        self._make_constraint = ConstraintMaker(self)

    @property
    def name(self) -> str:
        """The name of this component."""
        return "life_cycle_manager"

    @property
    def current_state(self) -> str:
        """The name of the current life cycle state."""
        return self._current_state.name

    def add_phase(self, phase_name: str, states: List[str], loop: bool = False):
        """Add a new phase to the lifecycle.

        Phases must be added in order.

        Parameters
        ----------
        phase_name
            The name of the phase to add.  Phase names must be unique.
        states
            The list of names (in order) of the states that make up the
            life cycle phase.  State names must be unique across the entire
            life cycle.
        loop
            Whether the life cycle phase states loop.

        Raises
        ------
        LifeCycleError
            If the phase or state names are non-unique.

        """
        self.lifecycle.add_phase(phase_name, states, loop)

    def set_state(self, state: str):
        """Sets the current life cycle state to the provided state.

        Parameters
        ----------
        state
            The name of the state to set.

        Raises
        ------
        LifeCycleError
            If the requested state doesn't exist in the life cycle.
        InvalidTransitionError
            If setting the provided state represents an invalid life cycle
            transition.

        """
        new_state = self.lifecycle.get_state(state)
        if self._current_state.valid_next_state(new_state):
            new_state.enter()
            self._current_state = new_state
        else:
            raise InvalidTransitionError(
                f"Invalid transition from {self.current_state} "
                f"to {new_state.name} requested."
            )

    def get_state_names(self, phase: str) -> List[str]:
        """Gets all states in the phase in their order of execution.

        Parameters
        ----------
        phase
            The name of the phase to retrieve the states for.

        Returns
        -------
        List[str]
            A list of state names in order of execution.

        """
        return self.lifecycle.get_state_names(phase)

    def add_handlers(self, state_name: str, handlers: List[Callable]):
        """Registers a set of functions to be called during a life cycle state.

        This method does not apply any constraints, rather it is used
        to build up an execution order for introspection.

        Parameters
        ----------
        state_name
            The name of the state to register the handlers for.
        handlers
            A list of functions that will execute during the state.

        """
        s = self.lifecycle.get_state(state_name)
        s.add_handlers(handlers)

    def add_constraint(
        self, method: Callable, allow_during: List[str] = (), restrict_during: List[str] = ()
    ):
        """Constrains a function to be executable only during certain states.

        Parameters
        ----------
        method
            The method to add constraints to.
        allow_during
            An optional list of life cycle states in which the provided
            method is allowed to be called.
        restrict_during
            An optional list of life cycle states in which the provided
            method is restricted from being called.

        Raises
        ------
        ValueError
            If neither ``allow_during`` nor ``restrict_during`` are provided,
            or if both are provided.
        LifeCycleError
            If states provided as arguments are not in the life cycle.
        ConstraintError
            If a lifecycle constraint has already been applied to the provided
            method.

        """
        if allow_during and restrict_during or not (allow_during or restrict_during):
            raise ValueError(
                'Must provide exactly one of "allow_during" or "restrict_during".'
            )
        unknown_states = (
            set(allow_during).union(restrict_during).difference(self.lifecycle._state_names)
        )
        if unknown_states:
            raise LifeCycleError(
                f"Attempting to constrain {method} with "
                f"states not in the life cycle: {list(unknown_states)}."
            )
        if restrict_during:
            allow_during = [
                s for s in self.lifecycle._state_names if s not in restrict_during
            ]

        self._make_constraint(method, allow_during)

    def __repr__(self) -> str:
        return f"LifeCycleManager(state={self.current_state})"

    def __str__(self) -> str:
        return str(self.lifecycle)


class LifeCycleInterface:
    """Interface to the life cycle management system.

    The life cycle management system allows components to constrain
    methods so that they're only available during certain simulation
    life cycle states.

    """

    def __init__(self, manager: LifeCycleManager):
        self._manager = manager

    def add_handlers(self, state: str, handlers: List[Callable]):
        """Registers a set of functions to be called during a life cycle state.

        This method does not apply any constraints, rather it is used
        to build up an execution order for introspection.

        Parameters
        ----------
        state
            The name of the state to register the handlers for.
        handlers
            A list of functions that will execute during the state.

        """
        self._manager.add_handlers(state, handlers)

    def add_constraint(
        self, method: Callable, allow_during: List[str] = (), restrict_during: List[str] = ()
    ):
        """Constrains a function to be executable only during certain states.

        Parameters
        ----------
        method
            The method to add constraints to.
        allow_during
            An optional list of life cycle states in which the provided
            method is allowed to be called.
        restrict_during
            An optional list of life cycle states in which the provided
            method is restricted from being called.

        Raises
        ------
        ValueError
            If neither ``allow_during`` nor ``restrict_during`` are provided,
            or if both are provided.
        LifeCycleError
            If states provided as arguments are not in the life cycle.
        ConstraintError
            If a life cycle constraint has already been applied to the
            provided method.

        """
        self._manager.add_constraint(method, allow_during, restrict_during)



import itertools
import math

import pandas as pd
import pytest

from vivarium.framework.population import (
    InitializerComponentSet,
    PopulationError,
    PopulationManager,
)

COL_NAMES = ["color", "count", "pie", "pi", "tracked"]
COLORS = ["red", "green", "yellow"]
COUNTS = [10, 20, 30]
PIES = ["apple", "chocolate", "pecan"]
PIS = [math.pi**i for i in range(1, 4)]
TRACKED_STATUSES = [True, False]
RECORDS = [
    (color, count, pie, pi, ts)
    for color, count, pie, pi, ts in itertools.product(
        COLORS, COUNTS, PIES, PIS, TRACKED_STATUSES
    )
]


@pytest.fixture(scope="function")
def population_manager():
    class _PopulationManager(PopulationManager):
        def __init__(self):
            super().__init__()
            self._population = pd.DataFrame(
                data=RECORDS,
                columns=COL_NAMES,
            )

        def _add_constraint(self, *args, **kwargs):
            pass

    return _PopulationManager()


def test_make_population_view(population_manager):
    pv = population_manager.get_view(COL_NAMES)
    assert pv._id == 0
    assert pv.name == "population_view_0"
    assert set(pv.columns) == set(COL_NAMES)
    assert pv.query is None

    # Failure here is lazy.  The manager should give you back views for
    # columns that don't exist since views are built during setup when
    # we don't necessarily know all the columns yet.
    cols = ["age", "sex", "tracked"]
    pv = population_manager.get_view(cols)
    assert pv._id == 1
    assert pv.name == "population_view_1"
    assert set(pv.columns) == set(cols)
    assert pv.query is None

    col_subset = ["color", "count"]
    pv = population_manager.get_view(col_subset)
    assert pv._id == 2
    assert pv.name == "population_view_2"
    assert set(pv.columns) == set(col_subset)
    # View will filter to tracked by default if it's not requested as a column
    assert pv.query == "tracked == True"

    q_string = "color == 'red'"
    pv = population_manager.get_view(COL_NAMES, query=q_string)
    assert pv._id == 3
    assert pv.name == "population_view_3"
    assert set(pv.columns) == set(COL_NAMES)
    assert pv.query == q_string


def test_population_view_subview(population_manager):
    pv = population_manager.get_view(COL_NAMES)

    col_subset = ["color", "count"]
    sub_pv = pv.subview(col_subset)
    assert set(sub_pv.columns) == set(col_subset)
    assert sub_pv.query == "tracked == True"

    col_subset = ["color", "count", "tracked"]
    sub_pv = pv.subview(col_subset)
    assert set(sub_pv.columns) == set(col_subset)
    assert sub_pv.query == pv.query

    col_subset = ["age", "sex"]
    with pytest.raises(PopulationError):
        pv.subview(col_subset)

    col_subset = COL_NAMES + ["age"]
    with pytest.raises(PopulationError):
        pv.subview(col_subset)

    col_subset = ["color", "count", "age", "sex"]
    with pytest.raises(PopulationError):
        pv.subview(col_subset)


def test_population_view_get(population_manager):
    pv = population_manager.get_view(COL_NAMES)
    full_idx = pd.RangeIndex(0, len(RECORDS))

    pop = pv.get(full_idx)
    assert set(pop.columns) == set(COL_NAMES)
    assert len(pop) == len(RECORDS)

    pop = pv.get(full_idx, query=f"color == 'red'")
    assert set(pop.columns) == set(COL_NAMES)
    assert len(pop) == len(RECORDS) // len(COLORS)

    cols_without_tracked = COL_NAMES[:-1]
    pv = population_manager.get_view(cols_without_tracked)

    pop = pv.get(full_idx)
    assert set(pop.columns) == set(cols_without_tracked)
    assert len(pop) == len(RECORDS) // 2

    pop = pv.get(full_idx, query=f"color == 'red'")
    assert set(pop.columns) == set(cols_without_tracked)
    assert len(pop) == len(RECORDS) // (2 * len(COLORS))

    sub_pv = pv.subview(cols_without_tracked)

    pop = sub_pv.get(full_idx)
    assert set(pop.columns) == set(cols_without_tracked)
    assert len(pop) == len(RECORDS) // 2

    pop = sub_pv.get(full_idx, query=f"color == 'red'")
    assert set(pop.columns) == set(cols_without_tracked)
    assert len(pop) == len(RECORDS) // (2 * len(COLORS))


def test_population_view_get_empty_idx(population_manager):
    pv = population_manager.get_view(COL_NAMES)

    pop = pv.get(pd.Index([]))
    assert isinstance(pop, pd.DataFrame)
    assert set(pop.columns) == set(COL_NAMES)
    assert pop.empty


def test_population_view_get_fail(population_manager):
    bad_pvs = [
        population_manager.get_view(["age", "sex"]),
        population_manager.get_view(COL_NAMES + ["age", "sex"]),
        population_manager.get_view(["age", "sex", "tracked"]),
        population_manager.get_view(["age", "sex"]),
        population_manager.get_view(["color", "count", "age"]),
    ]

    full_idx = pd.RangeIndex(0, len(RECORDS))

    for pv in bad_pvs:
        with pytest.raises(PopulationError):
            pv.get(full_idx)


def test_initializer_set_fail_type():
    component_set = InitializerComponentSet()

    with pytest.raises(TypeError):
        component_set.add(lambda: "test", ["test_column"])

    def initializer():
        return "test"

    with pytest.raises(TypeError):
        component_set.add(initializer, ["test_column"])


class UnnamedComponent:
    def initializer(self):
        return "test"


class Component:
    def __init__(self, name):
        self.name = name

    def initializer(self):
        return "test"

    def other_initializer(self):
        return "whoops"


def test_initializer_set_fail_attr():
    component_set = InitializerComponentSet()

    with pytest.raises(AttributeError):
        component_set.add(UnnamedComponent().initializer, ["test_column"])


def test_initializer_set_duplicate_component():
    component_set = InitializerComponentSet()
    component = Component("test")

    component_set.add(component.initializer, ["test_column1"])
    with pytest.raises(PopulationError, match="multiple population initializers"):
        component_set.add(component.other_initializer, ["test_column2"])


def test_initializer_set_duplicate_columns():
    component_set = InitializerComponentSet()
    component1 = Component("test1")
    component2 = Component("test2")
    columns = ["test_column"]

    component_set.add(component1.initializer, columns)
    with pytest.raises(PopulationError, match="both registered initializers"):
        component_set.add(component2.initializer, columns)

    with pytest.raises(PopulationError, match="both registered initializers"):
        component_set.add(component2.initializer, ["sneaky_column"] + columns)


def test_initializer_set():
    component_set = InitializerComponentSet()
    for i in range(10):
        component = Component(i)
        columns = [f"test_column_{i}_{j}" for j in range(5)]
        component_set.add(component.initializer, columns)


def test_get_view_with_no_query():
    manager = PopulationManager()
    view = manager._get_view(columns=["age", "sex"])
    assert view.query == "tracked == True"



import numpy as np
import pandas as pd
import pytest

from vivarium import InteractiveContext
from vivarium.framework.lookup import LookupTable, validate_parameters
from vivarium.testing_utilities import TestPopulation, build_table


@pytest.mark.skip(reason="only order 0 interpolation with age bin edges currently supported")
def test_interpolated_tables(base_config):
    year_start = base_config.time.start.year
    year_end = base_config.time.end.year
    years = build_table(lambda age, sex, year: year, year_start, year_end)
    ages = build_table(lambda age, sex, year: age, year_start, year_end)
    one_d_age = ages.copy()
    del one_d_age["year"]
    one_d_age = one_d_age.drop_duplicates()
    base_config.update(
        {"population": {"population_size": 10000}, "interpolation": {"order": 1}}
    )  # the results we're checking later assume interp order 1

    simulation = InteractiveContext(components=[TestPopulation()], configuration=base_config)
    manager = simulation._tables
    years = manager.build_table(
        years,
        key_columns=("sex",),
        parameter_columns=(
            "age",
            "year",
        ),
        value_columns=None,
    )
    ages = manager.build_table(
        ages,
        key_columns=("sex",),
        parameter_columns=(
            "age",
            "year",
        ),
        value_columns=None,
    )
    one_d_age = manager.build_table(
        one_d_age, key_columns=("sex",), parameter_columns=("age",), value_columns=None
    )

    pop = simulation.get_population(untracked=True)
    result_years = years(pop.index)
    result_ages = ages(pop.index)
    result_ages_1d = one_d_age(pop.index)

    fractional_year = simulation._clock.time.year
    fractional_year += simulation._clock.time.timetuple().tm_yday / 365.25

    assert np.allclose(result_years, fractional_year)
    assert np.allclose(result_ages, pop.age)
    assert np.allclose(result_ages_1d, pop.age)

    simulation._clock._time += pd.Timedelta(30.5 * 125, unit="D")
    simulation._population._population.age += 125 / 12

    result_years = years(pop.index)
    result_ages = ages(pop.index)
    result_ages_1d = one_d_age(pop.index)

    fractional_year = simulation._clock.time.year
    fractional_year += simulation._clock.time.timetuple().tm_yday / 365.25

    assert np.allclose(result_years, fractional_year)
    assert np.allclose(result_ages, pop.age)
    assert np.allclose(result_ages_1d, pop.age)


@pytest.mark.skip(reason="only order 0 interpolation with age bin edges currently supported")
def test_interpolated_tables_without_uninterpolated_columns(base_config):
    year_start = base_config.time.start.year
    year_end = base_config.time.end.year
    years = build_table(lambda age, sex, year: year, year_start, year_end)
    del years["sex"]
    years = years.drop_duplicates()
    base_config.update(
        {"population": {"population_size": 10000}, "interpolation": {"order": 1}}
    )  # the results we're checking later assume interp order 1

    simulation = InteractiveContext(components=[TestPopulation()], configuration=base_config)
    manager = simulation._tables
    years = manager.build_table(
        years,
        key_columns=(),
        parameter_columns=(
            "year",
            "age",
        ),
        value_columns=None,
    )

    result_years = years(simulation.get_population().index)

    fractional_year = simulation._clock.time.year
    fractional_year += simulation._clock.time.timetuple().tm_yday / 365.25

    assert np.allclose(result_years, fractional_year)

    simulation._clock._time += pd.Timedelta(30.5 * 125, unit="D")

    result_years = years(simulation.get_population().index)

    fractional_year = simulation._clock.time.year
    fractional_year += simulation._clock.time.timetuple().tm_yday / 365.25

    assert np.allclose(result_years, fractional_year)


def test_interpolated_tables__exact_values_at_input_points(base_config):
    year_start = base_config.time.start.year
    year_end = base_config.time.end.year
    years = build_table(lambda age, sex, year: year, year_start, year_end)
    input_years = years.year_start.unique()
    base_config.update({"population": {"population_size": 10000}})

    simulation = InteractiveContext(components=[TestPopulation()], configuration=base_config)
    manager = simulation._tables
    years = manager._build_table(
        years, key_columns=["sex"], parameter_columns=["age", "year"], value_columns=None
    )

    for year in input_years:
        simulation._clock._time = pd.Timestamp(year, 1, 1)
        assert np.allclose(
            years(simulation.get_population().index), simulation._clock.time.year + 1 / 365
        )


def test_lookup_table_scalar_from_list(base_config):
    simulation = InteractiveContext(components=[TestPopulation()], configuration=base_config)
    manager = simulation._tables
    table = manager._build_table(
        (1, 2), key_columns=None, parameter_columns=None, value_columns=["a", "b"]
    )(simulation.get_population().index)

    assert isinstance(table, pd.DataFrame)
    assert table.columns.values.tolist() == ["a", "b"]
    assert np.all(table.a == 1)
    assert np.all(table.b == 2)


def test_lookup_table_scalar_from_single_value(base_config):
    simulation = InteractiveContext(components=[TestPopulation()], configuration=base_config)
    manager = simulation._tables
    table = manager._build_table(
        1, key_columns=None, parameter_columns=None, value_columns=["a"]
    )(simulation.get_population().index)
    assert isinstance(table, pd.Series)
    assert np.all(table == 1)


def test_invalid_data_type_build_table(base_config):
    simulation = InteractiveContext(components=[TestPopulation()], configuration=base_config)
    manager = simulation._tables
    with pytest.raises(TypeError):
        manager._build_table(
            "break", key_columns=None, parameter_columns=None, value_columns=None
        )


def test_lookup_table_interpolated_return_types(base_config):
    year_start = base_config.time.start.year
    year_end = base_config.time.end.year
    data = build_table(lambda age, sex, year: year, year_start, year_end)

    simulation = InteractiveContext(components=[TestPopulation()], configuration=base_config)
    manager = simulation._tables
    table = manager._build_table(
        data, key_columns=["sex"], parameter_columns=["age", "year"], value_columns=None
    )(simulation.get_population().index)
    # make sure a single value column is returned as a series
    assert isinstance(table, pd.Series)

    # now add a second value column to make sure the result is a df
    data["value2"] = data.value
    table = manager._build_table(
        data, key_columns=["sex"], parameter_columns=["age", "year"], value_columns=None
    )(simulation.get_population().index)

    assert isinstance(table, pd.DataFrame)


@pytest.mark.parametrize(
    "data", [None, pd.DataFrame(), pd.DataFrame(columns=["a", "b", "c"]), [], tuple()]
)
def test_validate_parameters_no_data(data):
    with pytest.raises(ValueError, match="supply some data"):
        validate_parameters(data, [], [], [])


@pytest.mark.parametrize(
    "key_cols, param_cols, val_cols, match",
    [
        (None, None, None, "supply value_columns"),
        (None, None, [], "supply value_columns"),
        (None, None, ["a", "b"], "match the number of values"),
    ],
)
def test_validate_parameters_error_scalar_data(key_cols, param_cols, val_cols, match):
    with pytest.raises(ValueError, match=match):
        validate_parameters([1, 2, 3], key_cols, param_cols, val_cols)


@pytest.mark.parametrize(
    "key_cols, param_cols, val_cols, match",
    [(["a", "b"], ["b"], ["c"], "no overlap"), ([], ["b"], ["c"], "do not match")],
)
def test_validate_parameters_error_dataframe(key_cols, param_cols, val_cols, match):
    data = pd.DataFrame({"a": [1, 2], "b_start": [0, 5], "b_end": [5, 10], "c": [100, 150]})
    with pytest.raises(ValueError, match=match):
        validate_parameters(data, key_cols, param_cols, val_cols)


@pytest.mark.parametrize(
    "data", ["FAIL", pd.Interval(5, 10), "2019-05-17", {"a": 5, "b": 10}]
)
def test_validate_parameters_fail_other_data(data):
    with pytest.raises(TypeError, match="only allowable types"):
        validate_parameters(data, [], [], [])


@pytest.mark.parametrize(
    "key_cols, param_cols, val_cols",
    [
        (None, None, ["a", "b", "c"]),
        (None, ["d"], ["one", "two", "three"]),
        (["KEY"], None, ["a", "b", "c"]),
        (["KEY"], ["d"], ["a", "b", "c"]),
    ],
)
def test_validate_parameters_pass_scalar_data(key_cols, param_cols, val_cols):
    validate_parameters([1, 2, 3], key_cols, param_cols, val_cols)


@pytest.mark.parametrize(
    "key_cols, param_cols, val_cols",
    [(["a"], ["b"], ["c"]), ([], ["b"], ["c", "a"]), ([], ["b"], ["a", "c"])],
)
def test_validate_parameters_pass_dataframe(key_cols, param_cols, val_cols):
    data = pd.DataFrame({"a": [1, 2], "b_start": [0, 5], "b_end": [5, 10], "c": [100, 150]})
    validate_parameters(data, key_cols, param_cols, val_cols)


@pytest.mark.parametrize("validate", [True, False])
def test_validate_option_invalid_data(validate):
    if validate:
        with pytest.raises(ValueError, match="supply some data"):
            lookup = LookupTable(0, [], None, [], [], [], 0, None, True, validate)
    else:
        lookup = LookupTable(0, [], None, [], [], [], 0, None, True, validate)


@pytest.mark.parametrize("validate", [True, False])
def test_validate_option_valid_data(validate):
    data = [1, 2, 3]
    key_cols = ["KEY"]
    param_cols = ["d"]
    val_cols = ["a", "b", "c"]

    lookup = LookupTable(
        0, data, None, key_cols, param_cols, val_cols, 0, None, True, validate
    )



"""
=============
Interpolation
=============

Provides interpolation algorithms across tabular data for ``vivarium``
simulations.

"""
from typing import List, Tuple, Union

import numpy as np
import pandas as pd

ParameterType = Union[List[List[str]], List[Tuple[str, str, str]]]


class Interpolation:
    """A callable that returns the result of an interpolation function over input data.

    Attributes
    ----------
    data :
        The data from which to build the interpolation. Contains
        cateogrical_parameters and continuous_parameters.
    categorical_parameters :
        Column names to be used as categorical parameters in Interpolation
        to select between interpolation functions.
    continuous_parameters :
        Column names to be used as continuous parameters in Interpolation. If
        bin edges, should be of the form (column name used in call, column name
        for left bin edge, column name for right bin edge).
    order :
        Order of interpolation.
    """

    def __init__(
        self,
        data: pd.DataFrame,
        categorical_parameters: Union[List[str], Tuple[str]],
        continuous_parameters: ParameterType,
        order: int,
        extrapolate: bool,
        validate: bool,
    ):

        # TODO: allow for order 1 interpolation with binned edges
        if order != 0:
            raise NotImplementedError(
                f"Interpolation is only supported for order 0. You specified order {order}"
            )

        if validate:
            validate_parameters(data, categorical_parameters, continuous_parameters)

        self.key_columns = categorical_parameters
        self.data = data.copy()
        self.parameter_columns = continuous_parameters

        self.value_columns = self.data.columns.difference(
            set(self.key_columns) | set([col for p in self.parameter_columns for col in p])
        )

        self.order = order
        self.extrapolate = extrapolate
        self.validate = validate

        if self.key_columns:
            # Since there are key_columns we need to group the table by those
            # columns to get the sub-tables to fit
            sub_tables = self.data.groupby(list(self.key_columns))
        else:
            # There are no key columns so we will fit the whole table
            sub_tables = {None: self.data}.items()

        self.interpolations = {}

        for key, base_table in sub_tables:
            if (
                base_table.empty
            ):  # if one of the key columns is a category and not all values are present in data
                continue
            # since order 0, we can interpolate all values at once
            self.interpolations[key] = Order0Interp(
                base_table,
                self.parameter_columns,
                self.value_columns,
                self.extrapolate,
                self.validate,
            )

    def __call__(self, interpolants: pd.DataFrame) -> pd.DataFrame:
        """Get the interpolated results for the parameters in interpolants.

        Parameters
         ----------
        interpolants :
            Data frame containing the parameters to interpolate..

        Returns
        -------
        pd.DataFrame
            A table with the interpolated values for the given interpolants.
        """

        if self.validate:
            validate_call_data(interpolants, self.key_columns, self.parameter_columns)

        if self.key_columns:
            sub_tables = interpolants.groupby(list(self.key_columns))
        else:
            sub_tables = [(None, interpolants)]
        # specify some numeric type for columns so they won't be objects but will updated with whatever
        # column type actually is
        result = pd.DataFrame(
            index=interpolants.index, columns=self.value_columns, dtype=np.float64
        )
        for key, sub_table in sub_tables:
            if sub_table.empty:
                continue
            df = self.interpolations[key](sub_table)
            result.loc[sub_table.index, self.value_columns] = df.loc[
                sub_table.index, self.value_columns
            ]

        return result

    def __repr__(self):
        return "Interpolation()"


def validate_parameters(data, categorical_parameters, continuous_parameters):
    if data.empty:
        raise ValueError("You must supply non-empty data to create the interpolation.")

    if len(continuous_parameters) < 1:
        raise ValueError(
            "You must supply at least one continuous parameter over which to interpolate."
        )

    for p in continuous_parameters:
        if not isinstance(p, (List, Tuple)) or len(p) != 3:
            raise ValueError(
                f"Interpolation is only supported for binned data. You must specify a list or tuple "
                f"containing, in order, the column name used when interpolation is called, "
                f"the column name for the left edge (inclusive), and the column name for "
                f"the right edge (exclusive). You provided {p}."
            )

    # break out the individual columns from binned column name lists
    param_cols = [col for p in continuous_parameters for col in p]

    # These are the columns which the interpolation function will approximate
    value_columns = sorted(
        data.columns.difference(set(categorical_parameters) | set(param_cols))
    )
    if not value_columns:
        raise ValueError(
            f"No non-parameter data. Available columns: {data.columns}, "
            f"Parameter columns: {set(categorical_parameters)|set(continuous_parameters)}"
        )
    return value_columns


def validate_call_data(data, key_columns, parameter_columns):
    if not isinstance(data, pd.DataFrame):
        raise TypeError(
            f"Interpolations can only be called on pandas.DataFrames. You"
            f"passed {type(data)}."
        )
    callable_param_cols = [p[0] for p in parameter_columns]

    if not set(callable_param_cols) <= set(data.columns.values.tolist()):
        raise ValueError(
            f"The continuous parameter columns with which you built the Interpolation must all "
            f"be present in the data you call it on. The Interpolation has key "
            f"columns: {callable_param_cols} and your data has columns: "
            f"{data.columns.values.tolist()}"
        )

    if key_columns and not set(key_columns) <= set(data.columns.values.tolist()):
        raise ValueError(
            f"The key (categorical) columns with which you built the Interpolation must all"
            f"be present in the data you call it on. The Interpolation has key"
            f"columns: {key_columns} and your data has columns: "
            f"{data.columns.values.tolist()}"
        )


def check_data_complete(data, parameter_columns):
    """For any parameters specified with edges, make sure edges
    don't overlap and don't have any gaps. Assumes that edges are
    specified with ends and starts overlapping (but one exclusive and
    the other inclusive) so can check that end of previous == start
    of current.

    If multiple parameters, make sure all combinations of parameters
    are present in data.

    Requires that bins of each parameter be standard across all values
    of other parameters, i.e., all bins for one parameter when de-duplicated
    should cover a continuous range of that parameter with no overlaps or gaps
    and the range covered should be the same for all combinations of other
    parameter values.
    """

    param_edges = [
        p[1:] for p in parameter_columns if isinstance(p, (Tuple, List))
    ]  # strip out call column name

    # check no overlaps/gaps
    for p in param_edges:
        other_params = [p_ed[0] for p_ed in param_edges if p_ed != p]
        if other_params:
            sub_tables = data.groupby(list(other_params))
        else:
            sub_tables = {None: data}.items()

        n_p_total = len(set(data[p[0]]))

        for _, table in sub_tables:

            param_data = table[[p[0], p[1]]].copy().sort_values(by=p[0])
            start, end = param_data[p[0]].reset_index(drop=True), param_data[
                p[1]
            ].reset_index(drop=True)

            if len(set(start)) < n_p_total:
                raise ValueError(
                    f"You must provide a value for every combination of {parameter_columns}."
                )

            if len(start) <= 1:
                continue
            for i in range(1, len(start)):
                e = end[i - 1]
                s = start[i]
                if e > s or s == start[i - 1]:
                    raise ValueError(
                        f"Parameter data must not contain overlaps. Parameter {p} "
                        f"contains overlapping data."
                    )
                if e < s:
                    raise NotImplementedError(
                        f"Interpolation only supported for parameter columns "
                        f"with continuous bins. Parameter {p} contains "
                        f"non-continuous bins."
                    )


class Order0Interp:
    """A callable that returns the result of order 0 interpolation over input data.

    Attributes
    ----------
    data :
        The data from which to build the interpolation. Contains
        categorical_parameters and continuous_parameters.
    parameter_columns :
        Column names to be used as parameters in Interpolation.
    """

    def __init__(
        self,
        data,
        parameter_columns: ParameterType,
        value_columns: List[str],
        extrapolate: bool,
        validate: bool,
    ):
        """

        Parameters
        ----------
        data :
            Data frame used to build interpolation.
        parameter_columns :
            Parameter columns. Should be of form (column name used in call,
            column name for left bin edge, column name for right bin edge)
            or column name. Assumes left bin edges are inclusive and
            right exclusive.
        extrapolate :
            Whether or not to extrapolate beyond the edge of supplied bins.

        """
        if validate:
            check_data_complete(data, parameter_columns)

        self.data = data.copy()
        self.value_columns = value_columns
        self.extrapolate = extrapolate

        # (column name used in call, col name for left edge, col name for right):
        #               [ordered left edges of bins], max right edge (used when extrapolation not allowed)
        self.parameter_bins = {}

        for p in parameter_columns:
            left_edge = self.data[p[1]].drop_duplicates().sort_values()
            max_right = self.data[p[2]].drop_duplicates().max()

            self.parameter_bins[tuple(p)] = {
                "bins": left_edge.reset_index(drop=True),
                "max": max_right,
            }

    def __call__(self, interpolants: pd.DataFrame) -> pd.DataFrame:
        """Find the bins for each parameter for each interpolant in interpolants
        and return the values from data there.

        Parameters
        ----------
        interpolants:
            Data frame containing the parameters to interpolate..

        Returns
        -------
        pd.DataFrame
            A table with the interpolated values for the given interpolants.

        """
        # build a dataframe where we have the start of each parameter bin for each interpolant
        interpolant_bins = pd.DataFrame(index=interpolants.index)

        merge_cols = []
        for cols, d in self.parameter_bins.items():
            bins = d["bins"]
            max_right = d["max"]
            merge_cols.append(cols[1])
            interpolant_col = interpolants[cols[0]]
            if not self.extrapolate and (
                interpolant_col.min() < bins[0] or interpolant_col.max() >= max_right
            ):
                raise ValueError(
                    f"Extrapolation outside of bins used to set up interpolation is only allowed "
                    f"when explicitly set in creation of Interpolation. Extrapolation is currently "
                    f"off for this interpolation, and parameter {cols[0]} includes data outside of "
                    f"original bins."
                )
            bin_indices = np.digitize(interpolant_col, bins.tolist())
            # digitize uses 0 to indicate < min and len(bins) for > max so adjust to actual indices into bin_indices
            bin_indices[bin_indices > 0] -= 1
            interpolant_bins[cols[1]] = bins.loc[bin_indices].values

        index = interpolant_bins.index

        interp_vals = interpolant_bins.merge(self.data, how="left", on=merge_cols).set_index(
            index
        )
        return interp_vals[self.value_columns]



"""
=======================
Configuration Utilities
=======================

A set of functions for turning model specification files into programmatic
representations of :term:`model specifications <Model Specification>` and
:term:`configurations <Configuration>`.

"""
from pathlib import Path
from typing import Dict, Union

import yaml

from vivarium.config_tree import ConfigTree, ConfigurationError
from vivarium.framework.plugins import DEFAULT_PLUGINS


def build_model_specification(
    model_specification: Union[str, Path, ConfigTree] = None,
    component_configuration: Union[Dict, ConfigTree] = None,
    configuration: Union[Dict, ConfigTree] = None,
    plugin_configuration: Union[Dict, ConfigTree] = None,
) -> ConfigTree:
    if isinstance(model_specification, (str, Path)):
        validate_model_specification_file(model_specification)
        source = str(model_specification)
    else:
        source = "user_supplied_args"

    output_spec = _get_default_specification()
    output_spec.update(model_specification, layer="model_override", source=source)
    output_spec.components.update(
        component_configuration, layer="override", source="user_supplied_args"
    )
    output_spec.configuration.update(
        configuration, layer="override", source="user_supplied_args"
    )
    output_spec.plugins.update(
        plugin_configuration, layer="override", source="user_supplied_args"
    )

    return output_spec


def validate_model_specification_file(file_path: Union[str, Path]) -> None:
    """Ensures the provided file is a yaml file"""
    file_path = Path(file_path)
    if not file_path.exists():
        raise ConfigurationError(
            "If you provide a model specification file, it must be a file. "
            f"You provided {str(file_path)}",
            value_name=None,
        )

    if file_path.suffix not in [".yaml", ".yml"]:
        raise ConfigurationError(
            f"Model specification files must be in a yaml format. You provided {file_path.suffix}",
            value_name=None,
        )
    # Attempt to load
    with file_path.open() as f:
        raw_spec = yaml.full_load(f)
    top_keys = set(raw_spec.keys())
    valid_keys = {"plugins", "components", "configuration"}
    if not top_keys <= valid_keys:
        raise ConfigurationError(
            f"Model specification contains additional top level "
            f"keys {valid_keys.difference(top_keys)}.",
            value_name=None,
        )


def build_simulation_configuration() -> ConfigTree:
    return _get_default_specification().configuration


def _get_default_specification():
    default_config_layers = [
        "base",
        "user_configs",
        "component_configs",
        "model_override",
        "override",
    ]
    default_metadata = {"layer": "base", "source": "vivarium_defaults"}

    model_specification = ConfigTree(layers=default_config_layers)
    model_specification.update(DEFAULT_PLUGINS, **default_metadata)
    model_specification.update({"components": {}, "configuration": {}})

    user_config_path = Path("~/vivarium.yaml").expanduser()
    if user_config_path.exists():
        model_specification.configuration.update(user_config_path, layer="user_configs")

    return model_specification
